#include <FS.h>                  

#include <ESP8266WiFi.h>          //https://github.com/esp8266/Arduino

//needed for library
#include <DNSServer.h>
#include <ESP8266WebServer.h>
#include <WiFiManager.h>          //https://github.com/tzapu/WiFiManager

#include <ArduinoJson.h>          //https://github.com/bblanchon/ArduinoJson'

#include <PubSubClient.h>         

char mqtt_server[40] = "iot.eclipse.org";
char mqtt_port[6] = "1883";
char openfog_token[33] = "Your Username";
char static_ip[16] = "192.168.1.55";
char static_gw[16] = "192.168.1.1";
char static_sn[16] = "255.255.255.0";
char line_token[44] = "Line Token ( Option )";
char* MQTTtopic;

WiFiClient client;

void callback(char* topic, byte* payload, unsigned int length);
PubSubClient MQTTclient(mqtt_server, atoi(mqtt_port), callback, client);
bool shouldSaveConfig = false;

void saveConfigCallback () {
  Serial.println("Should save config");
  shouldSaveConfig = true;
}

#include <SoftwareSerial.h>
SoftwareSerial SoftSerial(14, 12, false, 256);  //D14 = RX, D12 = TX
bool MQTT_Connect_Flag = false;

const int Connection = 16;
long WiFiTimer_Cnt = 0;

String inputString = "";         
boolean stringComplete = false;  
word counter_read;
char RecieveString[100];
bool Recieve_Flag = false;

//แจ้งเตือนมีการบุกรุก
String message = "%E0%B9%81%E0%B8%88%E0%B9%89%E0%B8%87%E0%B9%80%E0%B8%95%E0%B8%B7%E0%B8%AD%E0%B8%99%E0%B8%81%E0%B8%B2%E0%B8%A3%E0%B8%9A%E0%B8%B8%E0%B8%81%E0%B8%A3%E0%B8%B8%E0%B8%81%20%E0%B8%9B%E0%B8%A3%E0%B8%B0%E0%B8%95%E0%B8%B9%E0%B8%AB%E0%B9%89%E0%B8%AD%E0%B8%87%E0%B8%97%E0%B8%B2%E0%B8%99%E0%B8%82%E0%B9%89%E0%B8%B2%E0%B8%A7%E0%B8%96%E0%B8%B9%E0%B8%81%E0%B9%80%E0%B8%9B%E0%B8%B4%E0%B8%94";

//=================================================================================================
void setup() {

        pinMode(Connection, OUTPUT);
        digitalWrite(Connection, HIGH);
  
        Serial.begin(115200);
        Serial.println();

        SoftSerial.begin(57600);
            
        if (SPIFFS.begin()) {
          if (SPIFFS.exists("/config.json")) {
            File configFile = SPIFFS.open("/config.json", "r");
            if (configFile) {
              size_t size = configFile.size();
              // Allocate a buffer to store contents of the file.
              std::unique_ptr<char[]> buf(new char[size]);
      
              configFile.readBytes(buf.get(), size);
              DynamicJsonBuffer jsonBuffer;
              JsonObject& json = jsonBuffer.parseObject(buf.get());
              json.printTo(Serial);
              if (json.success()) {
      
                strcpy(mqtt_server, json["mqtt_server"]);
                strcpy(mqtt_port, json["mqtt_port"]);
                strcpy(openfog_token, json["openfog_token"]);
                strcpy(line_token, json["line_token"]);
      
                if(json["ip"]) {
                  //static_ip = json["ip"];
                  strcpy(static_ip, json["ip"]);
                  strcpy(static_gw, json["gateway"]);
                  strcpy(static_sn, json["subnet"]);
                  Serial.println(static_ip);
                }
              } 
            }
          }
        } 

        WiFiManagerParameter custom_mqtt_server("server", "MQTT Server", mqtt_server, 40);
        WiFiManagerParameter custom_mqtt_port("port", "MQTT Port", mqtt_port, 5);
        WiFiManagerParameter custom_openfog_token("openfog", "Your username", openfog_token, 34);
        WiFiManagerParameter custom_line_token("line token", "Line Token ( Option )", line_token, 43);
      
        WiFiManager wifiManager;
        wifiManager.setSaveConfigCallback(saveConfigCallback);

        IPAddress _ip,_gw,_sn;
        _ip.fromString(static_ip);
        _gw.fromString(static_gw);
        _sn.fromString(static_sn);
      
        wifiManager.setSTAStaticIPConfig(_ip, _gw, _sn);
        
        wifiManager.addParameter(&custom_mqtt_server);
        wifiManager.addParameter(&custom_mqtt_port);
        wifiManager.addParameter(&custom_openfog_token);
        wifiManager.addParameter(&custom_line_token);
        
        wifiManager.setMinimumSignalQuality();
        
        if (!wifiManager.autoConnect("LineAlertBox", "password")) {       
          delay(3000);
          ESP.reset();
          delay(5000);
        }
   
        strcpy(mqtt_server, custom_mqtt_server.getValue());
        strcpy(mqtt_port, custom_mqtt_port.getValue());
        strcpy(openfog_token, custom_openfog_token.getValue());
        strcpy(line_token, custom_line_token.getValue());
      
        if (shouldSaveConfig) {
          DynamicJsonBuffer jsonBuffer;
          JsonObject& json = jsonBuffer.createObject();
          json["mqtt_server"] = mqtt_server;
          json["mqtt_port"] = mqtt_port;
          json["openfog_token"] = openfog_token;
          json["line_token"] = line_token;
      
          json["ip"] = WiFi.localIP().toString();
          json["gateway"] = WiFi.gatewayIP().toString();
          json["subnet"] = WiFi.subnetMask().toString();
      
          File configFile = SPIFFS.open("/config.json", "w");
          if (!configFile) {
            Serial.println("failed to open config file for writing");
          }
      
          json.prettyPrintTo(Serial);
          json.printTo(configFile);
          configFile.close();
          //end save
        }
      
        MQTTtopic = strcat(openfog_token,"/s1");
        reconnect();

        delay(2000);
        digitalWrite(Connection, LOW);
}

//Main Loop
//============================================================================
void loop() {
  
                if (!client.connected() && WiFi.status() == 3) {reconnect();}
               
                MQTTclient.loop();
            
                SoftSerialRead();
    
                if (stringComplete) {      
                    if(RecieveString[0] == '$' && RecieveString[1] == 'A' && RecieveString[2] == 'L' && RecieveString[3] == 'E' && RecieveString[4] == 'R' && RecieveString[5] == 'T')
                    {   
                          Line_Notify(message);
                    }    
                    stringComplete = false;
                }
    
                CheckWiFiConnection();
   
                delay(10); 
  
}
//==========================================================================
//End Main Loop

void SoftSerialRead()
{
    while (SoftSerial.available()) {
    char inChar = (char)SoftSerial.read();
    if(inChar == '$')
    {
      counter_read = 0;
      RecieveString[counter_read] = '$';
    }
    RecieveString[counter_read] = inChar;
    counter_read++;
    
    if (inChar == '*') {
      RecieveString[counter_read] = '\0';
      stringComplete = true;
    }
  }
}

//=========================================================================================================
void callback(char* topic, byte* payload, unsigned int length) {

        String topicStr = topic;  
        String PayloadString = String((char *)payload);
        SoftSerial.print(PayloadString);      

        WiFiTimer_Cnt = 0;
}

//=========================================================================================================
void CheckWiFiConnection()
{
  WiFiTimer_Cnt++;
  if(WiFiTimer_Cnt > 18000)    
  {
      WiFiTimer_Cnt = 0;
      if(WiFi.status() != WL_CONNECTED){  
        SoftSerial.print("$NOWIFI*");
      }

      if(WiFi.status() == WL_CONNECTED)
      {
        SoftSerial.print("$OKWIFI*");
      }
  }
  
}

//=========================================================================================================
void reconnect() {
  if(WiFi.status() != WL_CONNECTED){
    while (WiFi.status() != WL_CONNECTED) {
      delay(500);
      Serial.print(".");
    }
  }

  if(WiFi.status() == WL_CONNECTED){
    while (!client.connected()) {
      String clientName;
      clientName += "esp8266-";
      uint8_t mac[6];
      WiFi.macAddress(mac);
      clientName += macToStr(mac);

      if (MQTTclient.connect((char*) clientName.c_str())) {

        MQTTclient.subscribe(MQTTtopic);
        MQTT_Connect_Flag = true;
      }
      
      else{
        MQTT_Connect_Flag = false;
        abort();
      }
    }
  }
}
//=========================================================================================================

//generate unique name from MAC addr
String macToStr(const uint8_t* mac){

  String result;

  for (int i = 0; i < 6; ++i) {
    result += String(mac[i], 16);

    if (i < 5){
      result += ':';
    }
  }

  return result;
}
//=========================================================================================================
void Line_Notify(String message) {
   WiFiClientSecure client;

  if (!client.connect("notify-api.line.me", 443)) {
      return;   
  }

  String req = "";
  String reqAuthen = "";
  String req2 = "";
  req += "POST /api/notify HTTP/1.1\r\n";
  req += "Host: notify-api.line.me\r\n";
  reqAuthen += "Authorization: Bearer "; // + LineAccessToken + "\r\n";
  req2 += "Cache-Control: no-cache\r\n";
  req2 += "User-Agent: ESP8266\r\n";
  req2 += "Content-Type: application/x-www-form-urlencoded\r\n";
  req2 += "Content-Length: " + String(String("message=" + message).length()) + "\r\n";
  req2 += "\r\n";
  req2 += "message=" + message;
  
  client.print(req);
  client.print(reqAuthen);
  for(int i = 0; i<=42; i++)
  {
    client.print(line_token[i]);
  }
  client.print("\r\n");
  client.print(req2);
    
  delay(20);
  while(client.connected()) {
    String line = client.readStringUntil('\n');
    if (line == "\r") {
      break;
    }
  }
}

//===============================================================================================

